/* 
 * Copyright (C) 1999-2001 Peter T. Breuer <ptb@it.uc3m.es>
 */


#include <stdio.h>
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include "pidfile.h"

  static int
  openpidfile(struct nbd_pidfile *self, char * how) {

       self->pf = fopen(self->pidfile, how);

       if (!self->pf)
         return -errno;

       return 0;
  }

  static int
  closepidfile(struct nbd_pidfile *self) {
       int err;
       if (!self->pf)
         return -EINVAL;
       err = fclose(self->pf);
       self->pf = NULL;
       return err;
  }

  // PTB *re*make our pidfile with our pid inside after fork
  static int
  mymkpidfile(struct nbd_pidfile * self) {

    char c;
    int mypid = getpid();
    int err;

    // truncate
    err = openpidfile(self, "w+");
    if (err < 0)
        return err;

    err = fprintf(self->pf, "%-5u\n", mypid);
    if (err <= 0)
        goto error;

    // PTB fill with blanks to eof??
    while (fread(&c, 1, 1, self->pf) >= 1) {
      err = fseek(self->pf, -1, SEEK_CUR);
      if (err < 0)
        goto error;
      c = ' ';
      err = fwrite(&c, 1, 1, self->pf);
      if (err < 1)
        goto error;
    }

    closepidfile(self);
    return 0;

error:
    closepidfile(self);
    return -EIO;
  }

  // PTB *re*check our pidfile with our pid inside after fork
  static int
  readpidfile(struct nbd_pidfile * self) {

       int otherpid = 0;
       int err;

       err = openpidfile(self, "r");
       if (err < 0)
         return err;

       if (fscanf(self->pf, "%u\n", &otherpid) < 1) {
           closepidfile(self);
           return -EINVAL;
       }
       closepidfile(self);
       return otherpid;
  }

  static int
  unlockpidfile(struct nbd_pidfile *self) {
       int err;

       if (!self->pidfile)
         return -EINVAL;

       err = readpidfile(self);
       if (err < 0)
           return err;
       if (err != getpid())
           return -EINVAL;
       if (self->pf)
         closepidfile(self);
       err = unlink(self->pidfile);
       self->pf = NULL;
       return err;
  }

  // PTB *re*check our pidfile with our pid inside after fork
  static int
  myckpidfile(struct nbd_pidfile * self) {

       int otherpid;
       int mypid = getpid();
       int myppid = getppid();

       otherpid = readpidfile(self);

       if (otherpid == -1 || otherpid == mypid || otherpid == myppid)  {
           return 0;
       }
       if (kill (otherpid, 0) < 0) {
           return 0;
       }

       // PTB bad news, pidfile contains valid other pid
       return -EBUSY;
  }

  
  static int
  lockpidfile(struct nbd_pidfile * self) {
       int err;
       
       err = myckpidfile(self);
       if (err < 0) 
           return err; // failed to get lock
       err = mymkpidfile(self);
       return err;
  }


  int initpidfile(struct nbd_pidfile *self, char * pidfile) {

       self->pidfile = pidfile;
       self->read  = readpidfile;
       self->lock  = lockpidfile;
       self->unlock
                   = unlockpidfile;

       if (!self->pidfile)
         return -EINVAL;

       return 0;
  }

